"use strict"

class Board{
  constructor(div){
    // Board Condition parameters
    this.myRoleType = null // 0: sente as player, 1: gote as player, 2: watcher
    this.isPostGame = false
    this.studyHostType = null  // 0: not host, 1: Sub-host, 2: Host
    // Properties
    this._div = div
    this._originalWidth = div.width()
    this._originalHeight = div.height()
    this._komadais = new Array(2)
    this.playerInfos = new Array(2)
    this._timers = new Array(2)
    this.accumulatedTimes = new Array(2)
    this.endTime = null
    this._flags = new Array(2)
    this._publicPosition
    this._position
    this._direction = Position.CONST.SENTE
    this.pieceDesignType = 0
    this._selectedSquare = null
    this._lastSquare = null
    this._mouseDownSquare = null
    this._pickedPiece = null
    this._canMoveMyPiece = true
    this._canMoveHisPiece = true
    this.onListen = true
    this._initialPositionStr = null
    this.moves = new Array()
    this.game = null
    this.result = null // 0: sente win, 1: gote win, -1: draw
    this._kid = null
    this._rematchReady = [false, false]
    this._arrowsSelf = []
    this._arrowsPublic = []
    this._scale = 1
    this._promotionDialog = null
    this.hostMark = null
    this._generateParts()
    this._setTheme()
    this.loadPieceDesignOption()
  }

  _generateParts(){
    this._div.empty()
    this._ban = $('<div></div>', {id: 'banField'}).appendTo(this._div)
    this._coord = $('<div></div>', {id: 'coord'}).appendTo(this._div)
    this._komadais[0] = $('<div></div>', {id: 'senteKomadai', class: 'komadai'}).appendTo(this._div)
    this._komadais[1] = $('<div></div>', {id: 'goteKomadai', class: 'komadai'}).appendTo(this._div)
    this.playerInfos[0] = $('<div></div>', {id: 'senteInfo', class: 'player-info'}).appendTo(this._div)
    this.playerInfos[1] = $('<div></div>', {id: 'goteInfo', class: 'player-info'}).appendTo(this._div)
    this._timers[0] = new GameTimer($('<div></div>', {id: 'senteTimer', class: 'game-timer'}).appendTo(this._div))
    this._timers[1] = new GameTimer($('<div></div>', {id: 'goteTimer', class: 'game-timer'}).appendTo(this._div))
    this._flags[0] = $('<div></div>', {id: 'senteFlag', class: 'board-flag'}).appendTo(this._div)
    this._flags[1] = $('<div></div>', {id: 'goteFlag', class: 'board-flag'}).appendTo(this._div)
    for (let i = 0; i < 2; i++){
      this.playerInfos[i].append('<img class="mini-avatar"/>\
        <div><span id="player-info-mark">' + (i == 0 ? '☗' : '☖') + '</span>\
        <span id="player-info-name"></span><br><span id="player-info-rate"></span></div>')
    }
    $('div.player-info').click(function(){_playerNameDblClick($('span#player-info-name',$(this)).text())})
    $('div.player-info').each(function(){ tappablize($(this)) })
    this._arrowCanvas = $('<canvas></canvas>', {id: 'boardCanvas'}).attr({width: this._originalWidth, height: this._originalHeight}).appendTo(this._div)
    this._pickedPiece = $('<div></div>', {class: 'picked-piece'})
    this.hostMark = $('<i></i>', {class: 'fa fa-graduation-cap fa-lg'}).css('color', '#008')
    let thisInstance = this
    $('#senteKomadai, #goteKomadai').click(function(e){
      thisInstance._handleKomadaiClick()
    })
    $('#senteKomadai, #goteKomadai').each(function(){ tappablize($(this)) })
  }

  _setTheme(){
    let shrink = this._originalHeight / (1.3 * this._originalWidth) // have to schrink the board if height is not enough
    this._theme = "ichiji"
    this._banW = this._originalWidth * shrink * 0.9
    let zoom = this._banW / 410.0
    this._banH = 454.0 * zoom
    this._banX = (this._originalWidth - this._banW) / 2.0
    this._komaW = 43 * zoom
    this._komaH = 48 * zoom
    this._komadaiW = this._komaW * 7
    this._komadaiH = this._komaH * 1.3
    this._hisKomadaiX = 3 * zoom
    this._hisKomadaiY = 3 * zoom
    this._banY = this._hisKomadaiY + this._komadaiH + 3 * zoom
    this._myKomadaiX = this._originalWidth - this._komadaiW - 3 * zoom
    this._myKomadaiY = this._banY + this._banH + 3 * zoom
    this._myInfoX = this._myKomadaiX - this._originalWidth * 0.31 * shrink - 8 * zoom
    this._myInfoY = this._myKomadaiY
    this._myTimerX = this._myInfoX + 51 * zoom
    this._myTimerY = this._myKomadaiY + 32 * zoom
    this._myFlagX = this._myInfoX + 4 * zoom
    this._myFlagY = this._myTimerY
    this._hisTimerX = this._hisKomadaiX + this._komadaiW + 6 * zoom
    this._hisTimerY = this._hisKomadaiY
    this._hisInfoX = this._hisTimerX - 2 * zoom
    this._hisInfoY = this._hisTimerY + 32 * zoom
    this._hisFlagX = this._hisTimerX + 100 * zoom
    this._hisFlagY = 3 * zoom
    this._banEdgeX = 11 * zoom
    this._banEdgeY = 11 * zoom

    if (shrink < 1) $("div.player-info, div.game-timer, div.board-flag").css('transform', 'scale(' + shrink + ')')
    this._partSize()
    this._partLayout()

    this._preloadPieceImages(this._theme)
  }

  _partSize(){
    this._ban.css({width: this._banW + 'px', height: this._banH + 'px'})
    this._coord.css({width: this._banW + 'px', height: this._banH + 'px'})
    $(this._komadais.map(e => e[0])).css({width: this._komadaiW + 'px', height: this._komadaiH + 'px'})
    this._pickedPiece.css({width: this._komaW + 'px', height: this._komaH + 'px'})
  }

  _imagePath(){
    let dir = ['dobutsu', 'blind_extreme'].includes(this._theme) ? this._theme : 'default'
    this._ban.css('background-image', 'url(img/themes/' + dir + '/ban.jpg)')
    this._coord.css('background-image', 'url(img/themes/' + dir + (this._direction ? '/Scoord.png)' : '/Gcoord.png)'))
    this._komadais[this._direction ? 0 : 1].css('background-image', 'url(img/themes/' + dir + '/Shand.png)')
    this._komadais[this._direction ? 1 : 0].css('background-image', 'url(img/themes/' + dir + '/Ghand.png)')
  }

  _partLayout(){
    let myTurnIndex = this._direction ? 0 : 1
    let hisTurnIndex = this._direction ? 1 : 0
    this._ban.css({left: this._banX + 'px', top: this._banY + 'px'})
    this._coord.css({left: this._banX + 'px', top: this._banY + 'px'})
    this._komadais[myTurnIndex].css({left: this._myKomadaiX + 'px', top: this._myKomadaiY + 'px'})
    this._komadais[hisTurnIndex].css({left: this._hisKomadaiX + 'px', top: this._hisKomadaiY + 'px'})
    this.playerInfos[myTurnIndex].css({left: this._myInfoX + 'px', top: this._myInfoY + 'px'})
    this.playerInfos[hisTurnIndex].css({left: this._hisInfoX + 'px', top: this._hisInfoY + 'px'})
    this._flags[myTurnIndex].css({left: this._myFlagX + 'px', top: this._myFlagY + 'px'})
    this._flags[hisTurnIndex].css({left: this._hisFlagX + 'px', top: this._hisFlagY + 'px'})
    this._timers[myTurnIndex].setPosition(this._myTimerX, this._myTimerY)
    this._timers[hisTurnIndex].setPosition(this._hisTimerX, this._hisTimerY)
  }

  _generateSquares(){
    //Call after direction is set!
    this._ban.empty()
    for(let y = 1; y <= 9; y++){
      for(let x = 1; x <= 9; x++){
        let left
        let top
        if (this._direction) {
          left = this._banEdgeX + (9-x) * this._komaW
          top = this._banEdgeY + (y-1) * this._komaH
        } else {
          left = this._banEdgeX + (x-1) * this._komaW
          top = this._banEdgeY + (9-y) * this._komaH
        }
        let sq = $('<div></div>', {id: 'sq' + x + '_' + y, class: 'square'}).data({x: x, y: y})
        sq.css({width: this._komaW + 'px', height: this._komaH + 'px', left: left + 'px', top: top + 'px'})
        if (x < this.position.xmin || x > this.position.xmax || y < this.position.ymin || y > this.position.ymax) sq.addClass('square-wall')
        sq.appendTo(this._ban)
      }
    }
    let thisInstance = this
    $('.square').not('.square-wall').on("touchend", function(e){
      thisInstance._handleSquareTouchEnd(e)
    })
  }

  _refreshSquare(sq){
    sq.removeClass("square-last")
    if (sq.hasClass("square-wall")) return
    let koma = this._position.getPieceFromSquare(sq)
    if (koma) {
      sq.css('background-image', 'url(img/themes/' + this._theme + '/' + koma.toImagePath(!this._direction) + ')')
    } else {
      sq.css('background-image', 'none')
    }
  }

  _refreshPosition(){
    if (this._position == null) return
    let thisInstance = this
    for (let i = 0; i < 2; i++){
      this._komadais[i].empty()
    }
    $('.square').each(function(){
      thisInstance._refreshSquare($(this))
    })
    $(".square-selected").removeClass("square-selected")
//    $(".square-last").removeClass("square-last")
    for (let i = 0; i < 2; i++){
      let _layoutHandPiece = this._layoutHandPieceClosureFunc(i)
      this._position.komadais[i].sort(function(a,b){return a.type - b.type})
      this._position.komadais[i].forEach(function(piece){
        if (piece.CSA == 'OU' && this.game.gameType != 'vazoo') return
        let sq = $('<div></div>', {id: 'sq' + (piece.owner ? 0 : -1) + '_' + piece.getType(), class: 'square'}).data({x: piece.owner ? 0 : -1, y: piece.getType()})
        sq.css({width: this._komaW + 'px', height: this._komaH + 'px'})
        sq.css('background-image', 'url(img/themes/' + this._theme + '/' + piece.toImagePath(!this._direction) + ')')
        _layoutHandPiece(sq, piece)
        let thisInstance = this
        sq.on("touchend", function(e){
          thisInstance._handleSquareTouchEnd(e)
        })
        sq.appendTo(this._komadais[i])
      }, this)
      _layoutHandPiece = null
    }
    if (this._position.lastMove) {
      $('#sq' + this._position.lastMove.toX + '_' + this._position.lastMove.toY).addClass('square-last')
      $('#sq' + this._position.lastMove.fromX + '_' + this._position.lastMove.fromY).addClass('square-last')
    }
    if ($("#modalImpasse").dialog('isOpen')) this.calcImpasse()
    $(".name-popup").remove()
    this.highlightIfIllegal()
  }

  highlightIfIllegal(){
    let arrows = this.onListen ? this._arrowsPublic : this._arrowsSelf
    if (arrows.length > 0) return
    let obj = this.position.checkIllegal()
    if (obj){
      let name = EJ('ILLEGAL', '反則')
      if (obj.cause == 'NIFU'){
        this.addArrow(-1, obj.x1, obj.y1, obj.x1, obj.y1, 0xff0000, this.onListen, name)
        this.addArrow(-1, obj.x2, obj.y2, obj.x2, obj.y2, 0xff0000, this.onListen, name)
      } else if(obj.cause == 'SUICIDE'){
        this.addArrow(-1, obj.x1, obj.y1, obj.x2, obj.y2, 0xff0000, this.onListen, name)
      }
    }
  }

  _layoutHandPieceClosureFunc(i){
    let hash = this._position.handCoordinateHash(i) // This hash is used only to get hash.n
      // (using the same hash designed for desktop browser version)

    let direction = this._direction == (i == 0)
    let komadaiW = this._komadaiW
    let komadaiH = this._komadaiH
    let komaW = this._komaW
    let komaH = this._komaH
    return function(sq, piece){
      //square, piece
      let h = hash[piece.CSA]
      let sqX = direction ? komaW * (piece.type - 1) : komaW * (7 - piece.type)
      let sqY
      if (h.n == 1) {
        sqY = (komadaiH - komaH) / 2.0
      } else if (h.n == 2) {
        sqY = (komadaiH - komaH) * (0.8 - 0.6 * h.i)
      } else {
        sqY = (komadaiH - komaH) * (1 - h.i / (h.n - 1))
      }
      sq.css({left: sqX + 'px', top: sqY + 'px'})
      h.i += 1
    }
  }

  loadPieceDesignOption(){
    let newTheme
    let v = options.piece_type || 0
    if (v <= 8) newTheme = ['ichiji', 'ninju', 'hidetchi', 'ichiji_ryoko', 'dobutsu', 'kinki', 'ryoko', 'kiyoyasu', 'shogicz'][v]
    else if (v >= 100) newTheme = ['blind_middle', 'blind_hard', 'blind_extreme'][v - 100]
    if (this._theme != newTheme) this._preloadPieceImages(newTheme)
    this._theme = newTheme
    this._imagePath()
    this._refreshPosition()
  }

  _preloadPieceImages(theme){
    ['ou', 'gyoku', 'ryu', 'hi', 'uma', 'kaku', 'kin', 'ngin', 'gin', 'nkei', 'kei', 'nkyo', 'kyo', 'to', 'fu'].forEach(function(p){
      $('<img>').attr('src', 'img/themes/' + theme + '/S' + p + '.png')
      $('<img>').attr('src', 'img/themes/' + theme + '/G' + p + '.png')
    })
  }

  setScale(scale){
    if (scale == this._scale) return false // Not changed
    this._div.css('transform', 'scale(' + scale + ')')
    let hMargin = (scale - 1) * this._originalHeight / 2
    let wMargin = (scale - 1) * this._originalWidth / 2
    this._div.css('margin', hMargin + 'px ' + wMargin + 'px')
    this._scale = scale
    return true // Changed
  }

  actualWidth(){
    return this._div.width() * this._scale
  }

  actualHeight(){
    return this._div.height() * this._scale
  }

  setGame(game){
    this.game = game
    this.displayPlayerInfos()
    if (!this.game.isStudy()) {
      if (!this.game.isBlackIn && this.game.black.name != me.name) this.playerNameClassChange(0, "name-left", true)
      if (!this.game.isWhiteIn && this.game.white.name != me.name) this.playerNameClassChange(1, "name-left", true)
    }
  }

  setDirection(direction){
    this._direction = direction
    this._partLayout()
    this._imagePath()
  }

  displayPlayerInfos(){
    for (let i = 0; i < 2; i++){
      let user = i == 0 ? this.game.black : this.game.white
      this.playerInfos[i].find("img.mini-avatar").attr("src", user.avatarURL())
      this.playerInfos[i].find("#player-info-name").html(user.name)
      this.playerInfos[i].find("#player-info-rate").html('R: ' + user.rate + ' (' + makeRankFromRating(user.rate) + ')')
      if (user.titleName() != "") this.playerInfos[i].find("#player-info-rate").append('<span style="color:white;font-weight:bold;background:crimson;padding:0 0.2em;margin-left:0.3em;cursor:default">' + user.titleTag() + '</span>')
      this._flags[i].html(user.country.flagImgTagMovie())
    }
  }

  flipBoard(){
    this._cancelSelect()
    this.setDirection(!this._direction)
    this._generateSquares()
    this._refreshPosition()
    this.redrawAllArrows(this.onListen)
  }

  loadNewPosition(str = Position.CONST.INITIAL_POSITION){
    this._publicPosition = new Position(this.game ? this.game.gameType : null)
    this._publicPosition.superior = this.game ? (!this.game.isHandicap() && this.game.black.rate > this.game.white.rate) : false
    this._publicPosition.loadFromString(str)
    this._position = new Position(this.game ? this.game.gameType : null)
    this._position.superior = this.game ? (!this.game.isHandicap() && this.game.black.rate > this.game.white.rate) : false
    this._position.loadFromString(str)
    this._initialPositionStr = str
    this._generateSquares()
    this._refreshPosition()
    this._scratchKifu()
  }

  _scratchKifu(){
    this.moves = new Array()
    let firstMove = new Movement()
    this.moves.push(firstMove)
    kifuGrid.clear()
    this.addMoveToKifuGrid(firstMove)
  }

  static get MOVEMENT_CONST(){
    return Movement.CONST
  }

  addMoveToKifuGrid(move, highlight = true){
    if (move.num < kifuGrid.rows().count()) {
      while (kifuGrid.row(move.num).length == 1) kifuGrid.row(move.num).remove()
    }
    kifuGrid.row.add(move)
    kifuGrid.draw()
    if (highlight) {
      kifuGrid.rows().deselect()
      kifuGrid.row(move.num).select()
      scrollGridToSelected(kifuGrid)
    }
  }

  startGame(positionStr, myRoleType, move_strings = [], since_last_move = 0) {
    this.myRoleType = myRoleType
    if (myRoleType == 2) this.setDirection(true)
    else this.setDirection(myRoleType == 0)
    this.loadNewPosition(positionStr)
    this.clearArrows(true)
    this.clearArrows(false)
    this.isPostGame = false
    if (this.studyHostType == null) this.studyHostType = 0 // Keep the same host level as before even if startGame() is newly called
    this.onListen = true
    this._timers[0].initialize(this.game.total, this.game.byoyomi)
    this._timers[1].initialize(this.game.total, this.game.byoyomi)
    this.accumulatedTimes[0] = 0
    this.accumulatedTimes[1] = 0
    if (this.isPlayer()) this._timers[this.myRoleType].myPlayingTimer = true
    this.clearRematch()
    /*
    _board_coord_image.visible = false;
    studyOrigin = 0;
    */
    if (move_strings.length > 0) {
      move_strings.forEach(function(move_str){
        if (move_str.match(/^%TORYO/)) {
          resignTime = parseInt(move_str.split(",")[1])
          return
        }
        let move = this.getFinalMove().constructNextMove()
        move.setFromCSA(move_str.split(",")[0])
        move.setTime(parseInt(move_str.split(",")[1]), this)
        this.runningTimer.useTime(move.time)
        move = this._publicPosition.makeMove(move, false)
        this.moves.push(move)
        kifuGrid.row.add(move)
      }, this)
      this._position.deepCopy(this._publicPosition)
      this._refreshPosition()
      kifuGrid.draw()
      kifuGrid.rows().deselect()
      kifuGrid.row(":last").select()
      scrollGridToSelected(kifuGrid)
    }
		if (since_last_move > 0) {
      this.runningTimer.useTime(since_last_move, true)
		}
    this.runningTimer.run()
    this.updateTurnHighlight()
  }

  handleMonitorMove(move){
    this.runningTimer.useTime(move.time)
    move = this._publicPosition.makeMove(move)
    this.moves.push(move)
    if (!kifuGrid.row(':last').data().branch) {
      kifuGrid.row.add(move)
      if (this.onListen) {
        kifuGrid.draw()
        kifuGrid.rows().deselect()
        kifuGrid.row(":last").select()
        scrollGridToSelected(kifuGrid)
      } else {
        drawGridMaintainScroll(kifuGrid)
      }
    }
    if (this.onListen) {
      this._position.deepCopy(this._publicPosition)
      this._refreshPosition()
    }
    this.updateTurnHighlight()
    this.runningTimer.run()
  }

  refreshKifuGrid(){
  }

  _handleSquareMouseDown(sq){
    this._mouseDownSquare = sq
  }

  _handleSquareMouseUp(sq, circleEnabled){
    if (this._mouseDownSquare == null) return
    if (circleEnabled && this._mouseDownSquare.is(sq) || !this._mouseDownSquare.is(sq)) {
      if (this.onListen){
        if (!board.isPlaying()) {
          if (_studyBase != null || !this.isPostGame) this._addMyArrow(sq, true)
        }
      } else {
        this._addMyArrow(sq, false)
      }
    }
    this._mouseDownSquare = null
  }

  _handleSquareTouchEnd(e){
    let elem = $(e.currentTarget)
    let x = e.originalEvent.changedTouches[0].clientX
    let y = e.originalEvent.changedTouches[0].clientY
    let isTap = x >= elem.offset().left && x <= elem.offset().left + elem.width() && y >= elem.offset().top && y <= elem.offset().top + elem.height()
    if (isTap) {
      if (e.ctrlKey && elem.data().x > 0) {
        this._mouseDownSquare = elem
        this._handleSquareMouseUp(elem, getPremium() >= 1)
      } else {
        this._handleSquareClick(elem)
      }
    } else {
      let sqTo = this._getSquareFromMouseXY(x, y)
      if (sqTo) {
        if (this._selectedSquare) {
          this._handleSquareClick(sqTo)
        } else {
          this._mouseDownSquare = elem
          this._handleSquareMouseUp(sqTo)
        }
      }
    }
    e.preventDefault()
    e.stopPropagation()
  }

  _getSquareFromMouseXY(mouseX, mouseY){
    let banCornerX = this._div.offset().left + this._banX + this._banEdgeX
    let banCornerY = this._div.offset().top + this._banY + this._banEdgeY
    if (mouseX < banCornerX || mouseY < banCornerY) return null

    let sqI = null
    for (let i = 0; i < 9; i++){
      if (mouseX <= banCornerX + this._komaW * (i + 1)) {
        sqI = i
        break
      }
    }
    if (sqI == null) return null

    let sqJ = null
    for (let j = 0; j < 9; j++){
      if (mouseY <= banCornerY + this._komaH * (j + 1)) {
        sqJ = j
        break
      }
    }
    if (sqJ == null) return null

    if (this._direction) {
      return $("#sq" + (9-sqI) + "_" + (sqJ+1))
    } else {
      return $("#sq" + (sqI+1) + "_" + (9-sqJ))
    }
  }

  _addMyArrow(sqTo, isPublic){
    if (me && me.isGuest) return
    let fromType = -1
    let fromX = this._mouseDownSquare.data().x
    let fromY = this._mouseDownSquare.data().y
    let toX = sqTo.data().x
    let toY = sqTo.data().y
    if (fromX <= 0) {
      fromType = fromX == 0 ? 0 : 1
      fromX = fromY
      fromY = 0
      let thisInstance = this
      this._komadais[fromType].find('.square').each(function(i, node){
        if ($(node).is(thisInstance._mouseDownSquare)) {
          return false
        } else if ($(node).data().y == thisInstance._mouseDownSquare.data().y) {
          fromY++
        }
      })
    }
    this.addArrow(fromType, fromX, fromY, toX, toY, options.arrow_color, isPublic, me ? me.name : null)
    if (isPublic) client.gameChat("[##ARROW]" + fromType + "," + fromX + "," + fromY + "," + toX + "," + toY + ",0x" + options.arrow_color.toString(16))
  }

  _handleSquareClick(sq){
    if (!this._selectedSquare) {
      let koma = this._position.getPieceFromSquare(sq)
      if (this._canMovePieceNow() && koma && koma.owner == this._position.turn) {
        this._selectedSquare = sq
        sq.addClass("square-selected")
        if (!isTouchDevice && options.hold_piece && options.piece_type != 102) this._pickUpPiece(sq)
        if (this.isPlaying()) { //Send ##GRAB message
          let x = sq.data().x
          let y = sq.data().y
          if (x <= 0) {
            x = 100 + koma.type
            y = 0
          }
          sendGrab(x, y)
        }
        this._position.getMovableGridsFromSquare(sq).forEach(function(e){
          $("#sq" + e[0] + "_" + e[1]).addClass("square-movable")
          if (options.highlight_movable == 1 && options.piece_type != 102) $("#sq" + e[0] + "_" + e[1]).addClass("square-movable-highlight")
        })
      }
    } else {
      if (sq.hasClass("square-movable")) {
      	$(this._div).unbind("mouseleave")
        let res = this._position.canPromote(this._selectedSquare, sq)
        if (this.game.isKyoto() && this._selectedSquare.data().x <= 0) this._openPromotionDialog(sq, true) // Kyoto-shogi drop dialog
        else if (res == 2) this._manualMoveCommandComplete(sq, true)
        else if (res == 1) this._openPromotionDialog(sq)
        else this._manualMoveCommandComplete(sq, false)
      } else {
        this._cancelSelect()
      }
    }
  }

  _handleKomadaiClick(){
    if (this._selectedSquare) this._cancelSelect()
  }

  _pickUpPiece(sq){
    this._pickedPiece.css('background-image', sq.css('background-image'))
    this._positionPickedPiece(mouseX, mouseY)
    this._pickedPiece.appendTo(this._div)
    sq.addClass("square-picked-up")
    let thisInstance = this
  	$(this._div).mousemove(function(e){
      thisInstance._positionPickedPiece(e.pageX, e.pageY)
  	})
    $(this._div).mouseleave(function(e){
      if (e.pageX <= thisInstance._div.offset().left ||
          e.pageX >= thisInstance._div.offset().left + thisInstance.actualWidth() ||
          e.pageY <= thisInstance._div.offset().top ||
          e.pageY >= thisInstance._div.offset().top + thisInstance.actualHeight()) thisInstance._cancelSelect()
    })
  }

  _positionPickedPiece(mouse_X, mouse_Y){
    this._pickedPiece.css({
			top: (mouse_Y - this._div.offset().top) / this._scale - this._komaW/2 - 4 + 'px',
			left: (mouse_X - this._div.offset().left) / this._scale - this._komaH/2 + 2 + 'px'
		})
  }

  _openPromotionDialog(sq, kyotoFlipDrop = false){
    //square
    let thisInstance = this
    this._promotionDialog = $("<div></div>",{
      title: kyotoFlipDrop ? i18next.t("board.ask_flip") : i18next.t("board.ask_promote"),
      html: '<div id="promotion-window">\
        <span class="ui-helper-hidden-accessible"><input type="image"/></span>\
        <button type=button id="promote-yes" class="promotion-button"><img class="promotion-image" id="promote-yes-image"></button>\
        <button type=button id="promote-no" class="promotion-button"><img class="promotion-image" id="promote-no-image"></button>\
        </div>'
    }).dialog({
      dialogClass: 'no-close',
      modal: true,
      autoOpen: false,
      width: '70vw',
      position: {my:'top', at:'center', of:'#banField'},
      minHeight: 0,
      close: function(e){
        thisInstance._promotionDialog.dialog('destroy').remove()
        thisInstance._promotionDialog = null
      }
    })
    let koma = this._position.getPieceFromSquare(this._selectedSquare)
    this._promotionDialog.find('#promote-yes-image').attr('src', 'img/themes/' + this._theme + '/' + (kyotoFlipDrop ? koma.convertKyoto().toImagePath(!this._direction) : koma.toImagePath(!this._direction, true)))
    this._promotionDialog.find('#promote-no-image').attr('src', 'img/themes/' + this._theme + '/' + koma.toImagePath(!this._direction))
    $('#promote-yes').on('touchstart', function(){
      thisInstance._manualMoveCommandComplete(sq, true)
    })
    $('#promote-no').on('touchstart', function(){
      thisInstance._manualMoveCommandComplete(sq, false)
    })
    this._promotionDialog.dialog('open')
  }

  _manualMoveCommandComplete(destinationSquare, promote){
    //sqaure, boolean
    let move = kifuGrid.row({selected: true}).data().constructNextMove()
    move.setFromManualMove(this._position.turn, this._selectedSquare, destinationSquare, promote)
    this._cancelSelect()
    this._executeManualMove(move)
  }

  _executeManualMove(move){
    this.clearArrows(false)
    if (this.isPlaying()) {
      this.runningTimer.stop()
    }
    move = this._position.makeMove(move)
    if (this.isPlaying()) {
      sendMoveAsPlayer(move)
      this._publicPosition.deepCopy(this._position)
      this.updateTurnHighlight()
    } else {
      move.branch = true
      this.addMoveToKifuGrid(move)
      if (this.onListen && this.studyHostType >= 1) sendStudy()
      else forceKifuMode(0)
    }
    this._refreshPosition()
  }

  handleReceivedMove(move){
    move = this._publicPosition.makeMove(move)
    this.moves.push(move)
    this.addMoveToKifuGrid(move)
    if (this.onListen) {
      this._position.deepCopy(this._publicPosition)
      this._refreshPosition()
    }
  }

  replayMoves(moves){
    this._cancelSelect()
	  //_last_to_square = null;
    this._position.loadFromString(this._initialPositionStr)
    for (let i = 0; i < moves.length; i++) {
      if (moves[i].replayable()) this._position.makeMove(moves[i], false)
    }
    /*
		  if (mv.replayable()) {
			  _last_to_square = _cells[mv.to.y][mv.to.x];
			  _last_to_square.setStyle('backgroundColor', '0xCC3333');
			  if (mv.from.x < Kyokumen.HAND) {
				_last_from_square = _cells[mv.from.y][mv.from.x];
				_last_from_square.setStyle('backgroundColor', '0xFF5555');
			  }
		  }
      */
    this._refreshPosition()
  }

  handleBranchMove(move, withSound = false){
    return this._position.makeMove(move, withSound)
  }

  handleGrab(x, y){
    $(".square-selected").removeClass("square-selected")
    if (x == 0) {
      return
    } else if (x >= 100) {
      y = x - 100
      x = this._position.turn ? 0 : -1
    }
    $("#sq" + x + "_" + y).addClass("square-selected")
  }

  setResult(v){
    this.result = v // 0: sente win, 1: gote win, -1: draw
    if (v >= 0) {
    	board.playerNameClassChange(v, "name-winner", true)
    }
  }

  setBoardConditions(){
    if (this.isPlaying()) {
      this._canMoveMyPiece = true
      this._canMoveHisPiece = false
      this._coord.css('opacity', 0)
    } else {
      this._canMoveMyPiece = !this.onListen || this.studyHostType >= 1
      this._canMoveHisPiece = !this.onListen || this.studyHostType >= 1
      this._coord.css('opacity', 1)
    }
  }

  _canMovePieceNow(){
    return (this._canMoveMyPiece && this._position.turn == this._direction) || (this._canMoveHisPiece && this._position.turn != this._direction)
  }

  _cancelSelect(){
    if (this._promotionDialog) this._promotionDialog.dialog('close')
    if (this._selectedSquare != null){
    	$(this._div).unbind("mousemove mouseleave")
      $(".square").removeClass("square-movable square-movable-highlight")
      this._pickedPiece.remove()
      this._selectedSquare.removeClass("square-selected square-picked-up")
      this._selectedSquare = null
    }
    if (this.isPlaying()) sendGrab(0,0)
		//_endGrab();
		//_pieceGrab = false;
  }

  updateTurnHighlight(){
    if (this.isPostGame) {
      this.playerInfos[0].removeClass('player-info-has-turn')
      this.playerInfos[1].removeClass('player-info-has-turn')
    } else {
      this.playerInfos[this._publicPosition.turn ? 1 : 0].removeClass('player-info-has-turn')
      this.playerInfos[this._publicPosition.turn ? 0 : 1].addClass('player-info-has-turn')
    }
  }

  playerNameClassChange(turn, className, add_or_remove){
    //int, string, boolean
    if (add_or_remove) this.playerInfos[turn].find("#player-info-name").addClass(className)
    else this.playerInfos[turn].find("#player-info-name").removeClass(className)
  }

  pauseAllTimers(){
    this._timers[0].stop(true)
    this._timers[1].stop(true)
  }

  close(){
    this.myRoleType = null
    this.studyHostType = null
    this.game = null
    this.result = null
    this.playerInfos[0].find("#player-info-name").removeClass("name-winner name-left name-mouse-out")
    this.playerInfos[1].find("#player-info-name").removeClass("name-winner name-left name-mouse-out")
    this._timers[0].stop()
    this._timers[1].stop()
    this.endTime = null
  }

  rematch(turn){
    //integer
    this._rematchReady[turn] = true
  }

  rematchAgreed(){
    return this._rematchReady[0] == true && this._rematchReady[1] == true
  }

  clearRematch(){
    this._rematchReady = [false, false]
  }

  disconnectTimer(turn){
    //integer
    this._timers[turn].disconnect()
  }

  reconnectTimer(turn){
    //integer
    this._timers[turn].reconnect()
    if (this.isPlaying) this.runningTimer.run()
  }

  calcImpasse(){
    if (this.isPlaying() && this.myRoleType == (this._position.turn ? 0 : 1)) {
      _updateImpasseWindow(this._position.calcImpasse(), this.myRoleType)
    } else {
      _updateImpasseWindow(this._position.calcImpasse())
    }
  }

  refreshPosition(){
    this._refreshPosition()
  }

  addArrow(fromType, fromX, fromY, toX, toY, color, isPublic, sender){
    //integer, integer, integer, integer, integer, uint, boolean, string
    let arrows = isPublic ? this._arrowsPublic : this._arrowsSelf
    let isFull = arrows.length >= BoardArrow.CONST.MAX_ARROWS
    if (isFull) {
      arrows.shift()
      if (this.onListen == isPublic) this.redrawAllArrows(isPublic)
    }
    let arrow = new BoardArrow(fromType, fromX, fromY, toX, toY, color, sender, this._arrowCanvas)
    arrows.push(arrow)
    if (this.onListen == isPublic) {
      arrow.draw(this._scale)
      if (sender && me && sender != me.name) this.popupName(toX, toY, sender, color)
    }
  }

  clearArrows(isPublic, sender = "*"){
    let arrows = isPublic ? this._arrowsPublic : this._arrowsSelf
    let found = false
    if (sender != "*") {
      for (let i = arrows.length - 1; i >= 0; i--) {
        if (arrows[i].name == sender) {
          found = true
          arrows.splice(i,1)
        }
      }
    }
    if (!found) arrows.length = 0
    if (this.onListen && isPublic || !this.onListen && !isPublic) this.redrawAllArrows(isPublic)
    return found
  }

  redrawAllArrows(isPublic, withLabel = false){
    let arrows = isPublic ? this._arrowsPublic : this._arrowsSelf
    this.clearCanvas()
    arrows.forEach(function(arrow){
      arrow.draw(this._scale)
      if (withLabel) this.popupName(arrow.toX, arrow.toY, arrow.name, arrow.color)
    }, this)
  }

  popupName(x, y, name, color){
    let sq = $('#sq' + x + '_' + y)
    let posX = (sq.offset().left - this._div.offset().left) / this._scale + 0.56 * sq.width()
    let posY = (sq.offset().top - this._div.offset().top) / this._scale + 0.66 * sq.height()
    let element = $("<div></div>",{
      class: 'name-popup',
      html: '<span>' + name + '</span>'
    }).css({left: posX, top: posY, color: intToColorStyle(color)})
    this._div.append(element)
    element.delay(1500).animate({top: (posY - 0.3 * sq.height()) + 'px', opacity: 0}, 1500, function(){$(this).remove()})
  }

  clearCanvas(){
    this._arrowCanvas.get(0).getContext('2d').clearRect(0, 0, this._arrowCanvas.width(), this._arrowCanvas.height())
  }

  isPlayer(){
    return (this.myRoleType == 0 || this.myRoleType == 1)
  }

  isPlaying(){
    return this.isPlayer() && !this.isPostGame
  }

  isWatcher(){
    return this.myRoleType == 2
  }

  isHost(){
    return this.studyHostType == 2
  }

  isSubHost(){
    return this.studyHostType == 1
  }

  isPlayerPresent(turn){
    //integer: 0 or 1
    return !this.playerInfos[turn].find("#player-info-name").hasClass("name-left")
  }

  getPlayerRoleFromName(name){
    if (!this.game) return null
    if (name == this.game.black.name) return 0
    else if (name == this.game.white.name) return 1
    else return null
  }

  getFinalMove(){
    return this.moves[this.moves.length - 1]
  }

  getPlayersTimer(sente){
    return this._timers[sente ? 0 : 1]
  }

  getOpponent(){
    if (this.myRoleType == 0) return this.game.white
    else if (this.myRoleType == 1) return this.game.black
    else return null
  }

  setKifuId(kid){
    this._kid = kid
  }

  toKifuURL(withOption = false){
    if (this._kid == null || this._kid == 0) {
      return ""
    } else {
      let url = "http://system.81dojo.com/" + EJ('en', 'ja') + "/kifus/" + this._kid
      if (withOption) {
        let vars = []
        if (!this._direction) vars.push("turn=1")
        if (kifuGrid.row({selected: true}).index() != 0) vars.push("moves=" + kifuGrid.row({selected: true}).index())
      	if (vars.length > 0) url += "?" + vars.join("&")
      }
      return url
    }
  }

  getMaterialBalance(joban){
    return this._position.materialBalance(this._direction, joban)
  }

  get runningTimer(){
    return this._timers[this._publicPosition.turn ? 0 : 1]
  }

  get position(){
    return this._position
  }

  get div(){
    return this._div
  }

}
