"use strict"

class DobutsuBoard extends Board{
  _setTheme(){
    let shrink = this._originalHeight / (1.3 * this._originalWidth) // have to shrink the board if height is not enough
    this._theme = "3x4_pieco"
    this._banW = this._originalWidth * shrink * 0.98
    let zoom = this._banW / 410.0
    this._banH = 476 * zoom
    this._banX = (this._originalWidth - this._banW) / 2.0
    this._banY = 34 * zoom
    this._komadaiW = this._banW
    this._komadaiH = 76 * zoom
    this._hisKomadaiX = this._banX
    this._hisKomadaiY = this._banY + 3 * zoom
    this._myKomadaiX = this._banX
    this._myKomadaiY = this._banY + 398 * zoom
    this._myInfoX = this._banX + this._originalWidth * 0.17 * shrink
    this._myInfoY = this._banY + this._banH + 2 * zoom
    this._hisInfoX = this._myInfoX
    this._hisInfoY = 3 * zoom
    this._myTimerX = this._myInfoX + 145 * zoom
    this._myTimerY = this._myInfoY
    this._hisTimerX = this._myTimerX
    this._hisTimerY = this._hisInfoY
    this._myFlagX = this._myTimerX + 100 * zoom
    this._myFlagY = this._myInfoY
    this._hisFlagX = this._myFlagX
    this._hisFlagY = this._hisInfoY
    this._komaW = 75 * zoom
    this._komaH = 75 * zoom
    this._banEdgeX = 93 * zoom
    this._banEdgeY = 88 * zoom

    if (shrink < 1) $("div.player-info, div.game-timer, div.board-flag").css('transform', 'scale(' + shrink + ')')
    this._partSize()
    this._partLayout()

    this._preloadPieceImages(this._theme)
  }

  _imagePath(){
    this._ban.css('background-image', 'url(img/themes/' + this._theme + '/ban.png)')
    this._coord.css('background-image', 'url(img/themes/' + this._theme + (this._direction ? '/Scoord.png)' : '/Gcoord.png)'))
  }

  _generateSquares(){
    //Call after direction is set!
    this._ban.empty()
    for(let y = 1; y <= 4; y++){
      for(let x = 7; x <= 9; x++){
        let left
        let top
        if (this._direction) {
          left = this._banEdgeX + (9-x) * this._komaW
          top = this._banEdgeY + (y-1) * this._komaH
        } else {
          left = this._banEdgeX + (x-7) * this._komaW
          top = this._banEdgeY + (4-y) * this._komaH
        }
        let sq = $('<div></div>', {id: 'sq' + x + '_' + y, class: 'square'}).data({x: x, y: y})
        sq.css({width: this._komaW + 'px', height: this._komaH + 'px', left: left + 'px', top: top + 'px'})
        sq.appendTo(this._ban)
      }
    }
    let thisInstance = this
    $('.square').on("click", function(e){
      if (e.ctrlKey) return
      thisInstance._handleSquareClick($(this))
    })
    $('.square').on("mousedown", function(){
      thisInstance._handleSquareMouseDown($(this))
    })
    $('.square').on("mouseup", function(e){
      thisInstance._handleSquareMouseUp($(this), e.ctrlKey && getPremium() >= 1)
    })
  }

  _scratchKifu(){
    this.moves = new Array()
    let firstMove = new DobutsuMovement()
    this.moves.push(firstMove)
    kifuGrid.clear()
    this.addMoveToKifuGrid(firstMove)
  }

  static get MOVEMENT_CONST(){
    return DobutsuMovement.CONST
  }

  loadPieceDesignOption(){
    let v = options.piece_type_34 || 1
    let newTheme = ['3x4_pieco', '3x4_pieco', '3x4_hidetchi'][v]
    if (this._theme != newTheme) this._preloadPieceImages(newTheme)
    this._theme = newTheme
    this._imagePath()
    this._refreshPosition()
  }

  _preloadPieceImages(theme){
    ['hi', 'ki', 'ni', 'ra', 'zo'].forEach(function(p){
      $('<img>').attr('src', 'img/themes/' + theme + '/S' + p + '.png')
      $('<img>').attr('src', 'img/themes/' + theme + '/G' + p + '.png')
    })
  }

  _layoutHandPieceClosureFunc(i){
    let sum = this._position.komadais[i].length
    let deltaX = 0.9 * this._komaW
    let startX = (this._komadaiW - this._komaW - (sum - 1) * deltaX) * 0.5
    let count = 0
    return function(sq, piece){
      //square, piece
      let sqX = startX + deltaX * count
      sq.css({left: sqX + 'px', top: '0px'})
      count += 1
    }
  }

  highlightIfIllegal(){
    // Do nothing
  }
}
