"use strict";

class SwipeManager{
  constructor(){
    this._swipeOriginX = null
    this._swipeOriginY = null
    this._threshold = window.innerWidth * 0.08
    this._callbackFunctions = new Object()
  }

  _endTracking(){
    this._swipeOriginX = null
    this._swipeOriginY = null
  }

  registerTarget(elem, handler){
    // elem: jQuery object, handler: Callback function (it receives 'right', 'left', 'up', or 'down')
    let id = elem.prop('id')
    if (id == "") {
      console.log('Element has no id. Swipe handler cannot be registered.')
      return
    }
    this._callbackFunctions[id] = handler
    let _this = this
    elem.on("touchstart", function(e){
      _this._swipeOriginX = e.originalEvent.touches[0].clientX
      _this._swipeOriginY = e.originalEvent.touches[0].clientY
    })
    elem.on("touchmove", function(e){
      if (_this._swipeOriginX == null) return
      let clientX = e.originalEvent.touches[0].clientX
      let clientY = e.originalEvent.touches[0].clientY
      if (clientX - _this._swipeOriginX > _this._threshold) {
        _this._endTracking()
        _this._callbackFunctions[e.currentTarget.id]('right')
      } else if (clientX - _this._swipeOriginX < - _this._threshold) {
        _this._endTracking()
        _this._callbackFunctions[e.currentTarget.id]('left')
      } else if (clientY - _this._swipeOriginY > _this._threshold) {
        _this._endTracking()
        _this._callbackFunctions[e.currentTarget.id]('down')
      } else if (clientY - _this._swipeOriginY < - _this._threshold) {
        _this._endTracking()
        _this._callbackFunctions[e.currentTarget.id]('up')
      }
    })
    elem.on("touchend", function(e){
      _this._endTracking()
    })
  }
}
